import statsmodels.api as sm
import numpy as np
import pandas as pd

class pricing:
    def __init__(self, df):
        self.df = df

    def price_elasticity(self):
        df=self.df
        df = df[df["VOLUME"] > 0]
        df["Unit_price"] = df["SALES"]/df["VOLUME"]
        df_Stock = df.groupby(['VISIT_DT','SKU_ID']).agg({'Unit_price':'mean','VOLUME': 'mean' }).reset_index()
        x_pivot = df_Stock.pivot(index= 'VISIT_DT' ,columns='SKU_ID' ,values='Unit_price')
        x_values = pd.DataFrame(x_pivot.to_records())
        y_pivot = df_Stock.pivot( index = 'VISIT_DT',columns='SKU_ID', values='VOLUME')
        y_values = pd.DataFrame(y_pivot.to_records())
        points = []
        results_values = {
            "name": [],
            "price_elasticity": [],
            "price_mean": [],
            "quantity_mean": [],
            "intercept": [],
            "t_score":[],
            "slope": [],
            "coefficient_pvalue" : [],
        }
        #Append x_values with y_values per same product name
        for column in x_values.columns[1:]:
            column_points = []
            for i in range(len(x_values[column])):
                if not np.isnan(x_values[column][i]) and not np.isnan(y_values[column][i]):
                    column_points.append((x_values[column][i], y_values[column][i]))
            df_sam= pd.DataFrame(list(column_points), columns= ['x_value', 'y_value'])


            #Linear Regression Model
            x_value =df_sam['x_value']
            y_value = df_sam['y_value']
            X = sm.add_constant(x_value)
            model = sm.OLS(y_value, X)
            result = model.fit()
            #(Null Hypothesis test) Coefficient with a p value less than 0.05
            if result.f_pvalue < 0.05:

                rsquared = result.rsquared
                coefficient_pvalue = result.f_pvalue
                intercept,slope = result.params
                mean_price = np.mean(x_value)
                mean_quantity = np.mean(y_value)
                tintercept,t_score = result.tvalues

                #Price elasticity Formula
                price_elasticity = (slope)*(mean_price/mean_quantity)
                #price_elasticity = (slope)*(mean_quantity/mean_price)

                #Append results into dictionary for dataframe
                results_values["name"].append(column)
                results_values["price_elasticity"].append(price_elasticity)
                results_values["price_mean"].append(mean_price)
                results_values["quantity_mean"].append(mean_quantity)
                results_values["intercept"].append(intercept)
                results_values['t_score'].append(t_score)
                results_values["slope"].append(slope)
                results_values["coefficient_pvalue"].append(coefficient_pvalue)

        final_df = pd.DataFrame.from_dict(results_values)
        df_elasticity = final_df[['name','price_elasticity','t_score','coefficient_pvalue','slope','price_mean','quantity_mean','intercept']]
        Elastic_product= df_elasticity[df_elasticity['price_elasticity']<-1]
        Elastic_product_rank =  Elastic_product[['name', 'price_elasticity']]
        Elastic_product_rank = Elastic_product_rank.sort_values(by='price_elasticity')
        Elastic_product_rank['ranking'] = Elastic_product_rank['price_elasticity'].rank()
        Elastic_product_rank.reset_index(drop=True, inplace=True)

        UnitElastic_product= df_elasticity[df_elasticity['price_elasticity']>=-1]
        UnitElastic_product_rank =  UnitElastic_product[['name', 'price_elasticity']]
        UnitElastic_product_rank = UnitElastic_product_rank.sort_values(by='price_elasticity')
        UnitElastic_product_rank['ranking'] = UnitElastic_product_rank['price_elasticity'].rank()
        UnitElastic_product_rank.reset_index(drop=True, inplace=True)

        InElastic_product = df[['SKU_ID']]
        InElastic_product.rename(columns={'SKU_ID': 'name'}, inplace=True)
        # Use the ~ operator and isin method to filter out matching values
        InElastic_product = InElastic_product[~InElastic_product['name'].isin(df_elasticity['name'])]
        # Create a new DataFrame InElastic_product_rank with unique values in the 'name' column
        InElastic_product_rank = pd.DataFrame({'name': InElastic_product['name'].unique()})
        InElastic_product_rank.reset_index(drop=True, inplace=True)
        InElastic_product_rank.rename(columns={'index': 'Rank'}, inplace=True)
        return Elastic_product_rank,UnitElastic_product_rank,InElastic_product_rank


from business_models_initial import chi_performance, chi_performance_parsed_table
class feed_data:
  def __init__(self, data, percentage, column1, column2, column3):
    self.df=data
    self.percentage=percentage
    self.column1=column1
    self.column2=column2
    self.column3=column3
    obj=chi_performance.test(self.df,self.column1, self.column2, self.column3)
    a,b=obj.result()
    self.w=chi_performance_parsed_table.data_feed(self.percentage,self.df,a,b,self.column1, self.column2, self.column3)
  def interpret(self):
    w=self.w
    intprt=w.interpret()
    return intprt
  def output_table(self):
    w=self.w
    op_table=w.output_table()
    return op_table


import pandas as pd

class SLA_Determination:
    def __init__(self, df_selected, column1):
        self.df_selected = df_selected
        self.column1=column1

    def calculate_sla(self, st, et):
        df_sale = self.df_selected[(self.df_selected.saleflag == 1) & ((self.df_selected.order_date >= st) & (self.df_selected.order_date < et))]
        df_sale_perc = pd.DataFrame(data=df_sale[self.column1].quantile(q=[i/100 for i in range(75, 96)]))

        df_nosale = self.df_selected[(self.df_selected.saleflag == 0) & ((self.df_selected.lead_date >= st) & (self.df_selected.lead_date < et))]
        df_nosale_perc = pd.DataFrame(data=df_nosale[self.column1].quantile(q=[i/100 for i in range(75, 96)]))

        df_perc_diff = df_nosale_perc.Lead2FirstIntr_datedifference_Minute - df_sale_perc.Lead2FirstIntr_datedifference_Minute
        diff = pd.DataFrame(df_perc_diff)

        df_describe_sale = df_sale.describe()[[self.column1]].round(0)
        df_describe_sale = pd.concat([df_describe_sale.loc[['min', 'max', 'mean'], :], df_sale_perc.loc[[0.75, 0.80, 0.85, 0.90, 0.95], :]])
        df_describe_sale.index = ['Minimum', 'Maximum', 'Mean', '75th Perc', '80th Perc', '85th Perc', '90th Perc', '95th Perc']
        df_describe_sale.rename(columns={'Lead2FirstIntr_datedifference_Minute': 'Sale'}, inplace=True)
       
        df_describe_nosale = df_nosale.describe()[[self.column1]].round(0)
        df_describe_nosale = pd.concat([df_describe_nosale.loc[['min', 'max', 'mean'], :], df_nosale_perc.loc[[0.75, 0.80, 0.85, 0.90, 0.95], :]])
        df_describe_nosale.index = ['Minimum', 'Maximum', 'Mean', '75th Perc', '80th Perc', '85th Perc', '90th Perc', '95th Perc']
        df_describe_nosale.rename(columns={'Lead2FirstIntr_datedifference_Minute': 'Lead'}, inplace=True)

        df_table = pd.merge(df_describe_sale, df_describe_nosale, left_index=True, right_index=True)
        df_table = df_table.T.round(0)
        df_table.insert(0, 'Category', ['Sale', 'Non-Sale'])

        return df_sale_perc, df_nosale_perc, diff, df_table


from datetime import datetime as dt    
import numpy as np
import pandas as pd
from mlxtend.frequent_patterns import fpgrowth
from mlxtend.frequent_patterns import association_rules
from scipy.sparse import csr_matrix
from collections import defaultdict
import re

class recommandation:
    def __init__(self, df):
        self.df = df
        self.df1=self.recomendation_by_customer()
        self.df2=self.recomendation_by_product()
        self.df3=self.find_common_products()
        self.df4=self.mba()
        self.df5=self.convert_to_frozensets()
         
    def recomendation_by_customer (self):
        df=self.df
        df_crosstab = pd.crosstab(df['CUST_ID'], df['CAT_ID'])
        # if the value is more than 1 then replace it with 1
        df_crosstab[df_crosstab > 1] = 1
        # Do a total column wise
        df_crosstab['total_column'] = df_crosstab.sum(axis=1)
        # find the 90 percentile value
        ninety_percentile = np.percentile(df_crosstab['total_column'], 90)
        # keep only the top 90 percentile customers
        df_cat_crosstab_90 = df_crosstab[df_crosstab['total_column'] >= ninety_percentile]
        CT1 =df_crosstab.copy()
        CT2 =df_cat_crosstab_90.copy()
        jaccard_scores = {} # Creating an empty dictionary to store the Jaccard scores
        for cust_id1, row1 in CT1.iterrows(): # Creating a for loop to iterate through the rows of CT1
            jaccard_scores[cust_id1] = [] # Creating an empty list to store the Jaccard scores
            for cust_id2, row2 in CT2.iterrows(): # Creating a for loop to iterate through the rows of CT2
                intersection = row1 & row2 # Calculating the intersection of the rows
                union = row1 | row2 # Calculating the union of the rows
                jaccard_score = intersection.sum() / union.sum() # Calculating the Jaccard score
                jaccard_scores[cust_id1].append((cust_id2, jaccard_score)) # Appending the Jaccard score to the list
            jaccard_scores[cust_id1].sort(key=lambda x: x[1], reverse=True) # Sorting the list in descending order
            jaccard_scores[cust_id1] = [score[0] for score in jaccard_scores[cust_id1][:5]] # Selecting the top 5 customers

        n_dic=jaccard_scores.copy()
        cm_mat=df_crosstab.copy()

        prod = {} # Creating an empty dictionary to store the recommended products

        for i in n_dic.keys(): # Creating a for loop to iterate through the keys of n_dic
            prod[i] = () # Creating an empty tuple to store the recommended products

        for i in n_dic.keys(): # Creating a for loop to iterate through the keys of n_dic
            ncust = n_dic.get(i) # Storing the value of the key in ncust
            ncust.append(i) # Appending the key to ncust
            custprod_mat = cm_mat[cm_mat.index.isin(ncust)] # Creating a matrix with the customers in ncust
            custprod_mat.loc['coltotal'] = custprod_mat.sum() # Calculating the column total
            custprod_df = custprod_mat[custprod_mat.columns[custprod_mat.loc['coltotal'] != 0]] # Removing the columns with 0 values

            for b in custprod_df.index[~custprod_df.index.isin(['coltotal'])]: # Creating a for loop to iterate through the index of custprod_df
                c_b = custprod_df.loc[[b, 'coltotal']].T # Transposing the matrix
                c_b1 = c_b[c_b.iloc[:, 0] == 0] # Selecting the rows with 0 values
                c_b2 = np.round(c_b1[['coltotal']] / len(custprod_df), 3) # Calculating the percentage of 0 values
                c_l = list(c_b2.itertuples(name=None)) # Converting the dataframe to a list
                c_l = tuple(c_l) # Converting the list to a tuple
                prod[b] = prod[b] + c_l # Appending the tuple to the dictionary
                l1 = prod.get(b) # Storing the value of the key in l1
                l1_df = pd.DataFrame(l1, columns=['P_Name', 'rank']) # Creating a dataframe from the list
                l1_df1 = pd.DataFrame(l1_df.groupby('P_Name')['rank'].max()) # Calculating the maximum rank
                l1_df2 = l1_df1.sort_values(by=['rank'], ascending=False) # Sorting the dataframe in descending order
                list_of_tuples = l1_df2.to_records().tolist() # Converting the dataframe to a list of tuples
                tuple_of_tuples = tuple(list_of_tuples) # Converting the list to a tuple of tuples
                prod[b] = tuple_of_tuples # Storing the tuple of tuples in the dictionary

        recommended_products_dict=prod.copy()
        recomendation_customer = pd.DataFrame.from_dict(recommended_products_dict, orient='index')
        return recomendation_customer
   
    def recomendation_by_product (self):
        df=self.df
        p1=pd.crosstab(index=df['CAT_ID'],columns=df['CUST_ID'])   ## making a cross tab where rows are unique customers and columns are unique models
        p1[p1!=0]=1  ## substituting non zero values with 1 in the cross tab
        row_total=pd.DataFrame(p1.sum(axis=1),columns=['total'])  ## calculating sum for each row
        col_total=pd.DataFrame(p1.sum(axis=0),columns=['total'])  ## calculating sum for each column
        X = np.asmatrix(p1.values,dtype=np.float64) ## converting cross tab into a matrix
        sX = csr_matrix(X) ## converting X matrix into a sparse matrix
        numerator = sX.dot(sX.T)  ## calculating the numerator of the jaccared index i.e. the intersection part of the customers
        ones = np.ones(sX.shape[::-1],dtype=np.float64)  ## creating a unit matrix
        B = sX.dot(ones) ## multiplying the sparse matrix with unit matrix
        C=B+B.T ## Adding the matrix B & B transpose
        denominator=C-numerator ## calculating denominator of the jaccared index i.e. the union part of the customers
        j_matrix=numerator/denominator ## calculating jaccared index for each customer pair by dividing numerator and denominator
        j_matrix1=j_matrix.toarray() ## converting sparse matrix into a dense matrix : Scipy library change
        uppert_matrix=np.triu(j_matrix1,k=1)  ## considering only upper traingular matrix with diagonal elements 0.
        j_nonzero=pd.DataFrame(uppert_matrix[uppert_matrix > 0],columns=['j_index'])  ## considering only the non-zero elements of the upper traingular matrix
        j_matrix2=1-j_matrix1  ## calculating the distance for each pair of products
        j_matrix2[np.diag_indices_from(j_matrix1)] = np.nan  ## replacing the value of diagonal elements with 'nan' value of the distance matrix
        j_matrix1_df=pd.DataFrame(j_matrix2,index=p1.index,columns=p1.index)  ## converting distance matrix into a df and adding product ids
        x={}
        for i in j_matrix1_df.columns:

            u_df=j_matrix1_df[[i]]   ## getting the ith product column from the distance matrix
            u_df1=u_df[~ (u_df[i].isna())]   ## excluding the 'nan' value
            u_df1=u_df1[~ (u_df1[i]==1)]  ## excluding those products whose distance is 1 i.e. 100% from the ith product
            u_df2=u_df1.sort_values(by=[i])   ## sorting the distance values in ascending order
            u_df2=u_df2.reset_index()       ## resetting the df index
            x[i]=[tuple(r) for r in u_df2.to_numpy()]
        prodpair_list=x.copy()
        cust_list = p1.columns#[:1000]
        suggested_prod_list={}
        for i in cust_list:
       
            suggested_prod_list[i]=[]  ## creating keys in the empty dictionary for each custmer present in customer list
            m1=p1[[i]]    ## getting the i th customer column from the binary matrix  
            m_1=m1[m1[i]==1]   ## getting the product ids bought by ith customer
            purchased_prod=list(m_1.index)  ## making a list of product ids bought by ith customer
            filtered_d = dict((k, prodpair_list[k]) for k in purchased_prod if k in prodpair_list)  ## filtering the product neighbour dict according to purchased_prod list
            m_2=m1[m1[i]==0]  ## getting the product ids which are not bought by ith customer
            nonpurchased_prod=list(m_2.index)  ## making a list of product ids which are not bought by ith customer
       
            for j in purchased_prod:  ## generating a for loop on the purchased_prod list
                nprod=filtered_d.get(j)  ## filtering the filtered_d dict based on j th product and getting the neighbour list for j th prod
                df=pd.DataFrame.from_records(nprod, columns =['prod','distance'])  ## creating a df from the list of tuples
                df1=df[df['prod'].isin(nonpurchased_prod)].head(1)  ## filtering the above df based on nonpurchased product list of ith customer and capturing only the first non purchased product with its rank (i.e. distance) if there is any nonpurchased product from the neighbour list of j th purchased product of ith customer
                   
                if len(df1.index)!=0:   ## if the above df  is not empty then
                    tup1=[(k, v) for k, v in zip(df1['prod'],df1['distance'])][0]  ## make a tuple from the above df i.e. ('nonpurchased_prod_id','rank')
                    suggested_prod_list[i].append(tup1)  ## appending the tuple in the list of ith customer's recommended product list
        for k in suggested_prod_list.keys():
           
            l1=suggested_prod_list.get(k)  ## getting the recommended products for b th customer                 
            l1_df = pd.DataFrame(l1, columns =['P_Name', 'rank'])  ## converting the tuple of recommended products to a df      
            l1_df1=pd.DataFrame(l1_df.groupby('P_Name')['rank'].max())  ## groupby the df wrt unique products with highest rank   
            l1_df2=l1_df1.sort_values(by=['rank'],ascending=True)  ## sorting the unique recommended products according to ranks
            list_of_tuples = l1_df2.to_records().tolist()  ## converting the dataframe to a list of tuples
            suggested_prod_list[k]=list_of_tuples  ## final list of recommended products for b th customer
        df_rec_prod = pd.DataFrame.from_dict(suggested_prod_list, orient='index')
        # add a index label s as "CUST_ID"
        df_rec_prod.index.name = 'CUST_ID'
        return df_rec_prod
   
    def find_common_products(self):
        df1=self.df1
        df2=self.df2
        common_products = {} # Creating an empty dictionary to store the common products
        # Iterate over the row indexes (customer ids) of the first dataframe
        for customer_id in df1.index:
            # Check if the customer ID exists in both dataframes
            if customer_id in df2.index:
                # Get the products for the current customer from both dataframes
                products1 = set([item[0] for item in df1.loc[customer_id] if pd.notnull(item)]) # Creating a set of the products
                products2 = set([item[0] for item in df2.loc[customer_id] if pd.notnull(item)]) # Creating a set of the products
           
                # Find the common products for the current customer
                common_products[customer_id] = list(products1.intersection(products2)) # Creating a list of the common products
   
        # Create a new dataframe with the common products
        df_common = pd.DataFrame.from_dict(common_products, orient='index') # Creating a dataframe from the dictionary
        df_common.index.name = 'Customer' # Renaming the index column
        #df_common.set_index('Customer', inplace=True)
        # appenindg Recomendation to each column name
        
        df_common.columns = ['Recommendation_' + str(col) for col in df_common.columns] # Renaming the columns
        # change itesm to intgers
        df_common = df_common.astype('Int64') # Changing the datatype of the dataframe to integer
        df_common.reset_index(inplace=True)
        return df_common
    
    def mba(self):
        threshold = 0.01  # Adjust the minimum support threshold as needed
        df_original = self.df
        basket = df_original.groupby(['VISIT_ID', 'CAT_ID']).size().unstack().fillna(0)
        basket[basket != 0] = 1
        basket.max(axis=1).sum()
        minTransaction = len(basket.index) * threshold
        totalTransactions = len(basket.index)
        min_support_calc = minTransaction / totalTransactions

        # Perform frequent itemset mining
        frequent_itemsets = fpgrowth(basket, min_support=min_support_calc, use_colnames=True)
        frequent_itemsets.describe()

        # Adjust the minimum threshold for association rules to capture more rules
        rules = association_rules(frequent_itemsets, metric="lift", min_threshold=0.2)  # Adjust the threshold as needed
        rules.sort_values('confidence', ascending=False, inplace=True)
        
        return rules

    def convert_to_frozensets(self):
        dataframe=self.df4
        # Iterate over each column in the DataFrame
        for column in dataframe.columns:
            # Convert the column to a frozenset and update the DataFrame
            dataframe[column] = dataframe[column].apply(lambda x: frozenset(map(int, re.findall(r'\d+', str(x)))))
        return dataframe
    
    def final_recomendations(self):
        df=self.df
        df_common=self.df3
        basket_P = df.groupby(['CUST_ID','CAT_ID']).size().unstack().fillna(0)    
        basket_P[basket_P!=0]=1
        basket_P['purchases'] = basket_P.apply(lambda x: tuple(x.index[x==1]), axis=1)
        basket_P['purchases'] = basket_P['purchases'].apply(lambda x: frozenset(x))
        # drop all columns except purchases
        basket_P.drop(columns=basket_P.columns.difference(['purchases']), inplace=True)
        basket_P['recomendations'] = None
        for index, row in basket_P.iterrows():
            cust_id = index
            recomendations = df_common[df_common ['Customer'] == cust_id].iloc[:, 0:].values.tolist()
            recomendations = [item for sublist in recomendations for item in sublist]
            # remove the nan values
            recomendations = [x for x in recomendations if str(x) != 'nan']
            # make it int 
            recomendations = [int(x) if pd.notna(x) else 0 for x in recomendations]
            recomendations = frozenset(recomendations)
            basket_P.at[cust_id, 'recomendations'] = recomendations
        df_mba=self.df4
        recommendations = defaultdict(set)
        for _, row in basket_P.iterrows():
            purchases = row['purchases']
            for _, mba_row in df_mba.iterrows():
                antecedent = frozenset(mba_row['antecedents'])  # Convert antecedent to a frozenset
                consequent = mba_row['consequents']
                confidence = mba_row['confidence']
                if antecedent.issubset(purchases) and isinstance(confidence, (int, float)) and confidence >= 0:
                    recommendations[row.name].update(consequent)

        rec_mba_column = [recommendations.get(cust_id, set()) for cust_id in basket_P.index]
        basket_P['recommendation_mba'] = rec_mba_column
        # change the datatype to frozenset
        basket_P['recommendation_mba'] = basket_P['recommendation_mba'].apply(lambda x: frozenset(x))
        basket_P['recommendation_mba'] = basket_P.apply(lambda x: x['recommendation_mba'] - x['purchases'], axis=1)
        basket_P['common'] = None
        for index, row in basket_P.iterrows():
            common = row['recomendations'].intersection(row['recommendation_mba'])
            # if common is empty then set it to None
            if len(common) == 0:
                common = None
            basket_P.at[index, 'common'] = common
        return basket_P
    



import pandas as pd

class SLADetermination:
    def __init__(self, df):
        self.df = df
   
    def filter_and_convert_dates(self, st, et):
        selected_columns = ['Lead2FirstIntr_datedifference_Minute', 'lead_source', 'saleflag', 'salesman_id', 'order_date', 'order_term', 'net_amount', 'matched_table', 'lead_date', 'interactions']
        self.df = self.df[selected_columns]
       
        self.df = self.df[(self.df['saleflag'] == 1) & ((self.df['order_date'] >= st) & (self.df['order_date'] < et))]
       
        self.df['order_date'] = pd.to_datetime(self.df['order_date'])
        self.df['lead_date'] = pd.to_datetime(self.df['lead_date'])
   
    def calculate_lead_to_sale_diff(self):
        self.df['Lead_to_Sale_diff'] = (self.df['order_date'] - self.df['lead_date']).dt.days
        self.df = self.df[self.df['Lead_to_Sale_diff'] > 0]

    def categorize_lead_to_sale_calculate(self):
        bins = [7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18]
        labels = ['7 days', '8 days', '9 days', '10 days', '11 days', '12 days', '13 days', '14 days', '15 days', '16 days', '17 days']
   
        self.df['Lead_to_Sale_Group'] = pd.cut(self.df['Lead_to_Sale_diff'], bins=bins, labels=labels, right=False)
        self.df_sale_inter = self.df.copy()
       
        bins_int = [2, 3, 4, 5, 6, 7, 8, 9, 10]
        labels = ['2 Interactions', '3 Interactions', '4 Interactions', '5 Interactions', '6 Interactions', '7 Interactions', '8 Interactions', '9 Interactions']
       
        self.df_sale_inter['Lead_to_Sale_Interactions_Group'] = pd.cut(self.df_sale_inter['interactions'], bins=bins_int, labels=labels, right=False)

        self.df_sale_age_cum = self.df.groupby('Lead_to_Sale_Group').size().reset_index(name='Count')
        self.df_sale_age_cum['Cumulative_Count'] = self.df_sale_age_cum['Count'].cumsum()
        self.df_sale_age_cum['Cumulative_Percentage'] = (self.df_sale_age_cum['Count'].cumsum() / self.df_sale_age_cum['Count'].sum()) * 100

        self.df_sale_int_cum = self.df_sale_inter.groupby('Lead_to_Sale_Interactions_Group').size().reset_index(name='Count')
        self.df_sale_int_cum['Cumulative_Count'] = self.df_sale_int_cum['Count'].cumsum()
        self.df_sale_int_cum['Cumulative_Percentage'] = (self.df_sale_int_cum['Count'].cumsum() / self.df_sale_int_cum['Count'].sum()) * 100
   
        return self.df_sale_age_cum, self.df_sale_int_cum


import setuptools

with open("README.md", "r") as fh:
	long_description1 = fh.read()

setuptools.setup(
	# Here is the module name.
	name="businessmodels",

	# version of the module
	version="0.2.8",

	# Name of Author
	author="Business Brio",

	# your Email address
	author_email="gautam@businessbrio.com",

	# #Small Description about module
	description="It is a Data Science package containing many modules to help in analysis.",


	# Specifying that we are using markdown file for description
	long_description=long_description1,
	long_description_content_type="text/markdown",

	# Any link to reach this module, ***if*** you have any webpage or github profile
	url="https://github.com/Business-Brio/businessmodels/tree/main",
	packages=setuptools.find_packages(),


	# if module has dependencies i.e. if your package rely on other package at pypi.org
	# then you must add there, in order to download every requirement of package



	install_requires=["business-models-initial"],


	license="MIT",

	# classifiers like program is suitable for python3, just leave as it is.
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
)



from datetime import datetime

class QAmodels:
    def __init__(self, schema):
        self.schema = schema  # Store schema as an instance attribute

    def enhancement_prompt(self, user_question, prompt_instructions=""):
        """
        Enhances the user's query by adding relevant schema details, instructions, and the current date.
    
        IMPORTANT:
        - "Last quarter" without a year refers to the current quarter
        - "Last quarter of YYYY" refers to Q4 of FY YYYY-(YYYY+1), which is Jan-Mar of (YYYY+1)
        """
        # Use the current date
        current_date = datetime.now().strftime('%Y-%m-%d')
 
        # Extract year and month from current date
        current_datetime = datetime.now()
        current_year = current_datetime.year
        current_month = current_datetime.month
    
        # Determine the current financial year
        current_fy_start_year = current_year if current_month >= 4 else current_year - 1
        current_fy_end_year = current_fy_start_year + 1
    
        # Determine the current quarter within the financial year
        if 4 <= current_month <= 6:
            current_quarter = 1
        elif 7 <= current_month <= 9:
            current_quarter = 2
        elif 10 <= current_month <= 12:
            current_quarter = 3
        else:  # 1 <= current_month <= 3
            current_quarter = 4
    
        # Define exact date ranges for current and previous year quarters
        if current_quarter == 1:
            current_quarter_start = f"{current_fy_start_year}-04-01"
            current_quarter_end = f"{current_fy_start_year}-06-30"
            prev_year_quarter_start = f"{current_fy_start_year-1}-04-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-06-30"
        elif current_quarter == 2:
            current_quarter_start = f"{current_fy_start_year}-07-01"
            current_quarter_end = f"{current_fy_start_year}-09-30"
            prev_year_quarter_start = f"{current_fy_start_year-1}-07-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-09-30"
        elif current_quarter == 3:
            current_quarter_start = f"{current_fy_start_year}-10-01"
            current_quarter_end = f"{current_fy_start_year}-12-31"
            prev_year_quarter_start = f"{current_fy_start_year-1}-10-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-12-31"
        else:  # current_quarter == 4
            current_quarter_start = f"{current_fy_end_year}-01-01"
            current_quarter_end = f"{current_fy_end_year}-03-31"
            prev_year_quarter_start = f"{current_fy_end_year-1}-01-01"
            prev_year_quarter_end = f"{current_fy_end_year-1}-03-31"

        enhancement_prompt = f"""
    You are a highly skilled SQL expert and data analyst. Your task is to enhance a given user question
    by making it more structured, detailed, and aligned with the provided schema and prompt instructions.
   
    Today's Date: {current_date}
    Schema: {self.schema}
    Prompt Instructions: {prompt_instructions}
    User Question: {user_question}
   
    CRITICAL INFORMATION ABOUT DATE INTERPRETATIONS:
   
    The organization follows an April-March financial year structure:
    - Financial Year X-Y runs from April 1 of year X to March 31 of year Y
    - The current financial year is FY {current_fy_start_year}-{current_fy_end_year}
    - The current quarter is Q{current_quarter} of FY {current_fy_start_year}-{current_fy_end_year}
   
    The quarters in each financial year are defined as:
    - Q1: April 1 to June 30
    - Q2: July 1 to September 30
    - Q3: October 1 to December 31
    - Q4: January 1 to March 31
   
    EXPLICIT DATE MAPPINGS FOR TODAY ({current_date}):
    - "Last quarter" or "current quarter" refers to: {current_quarter_start} to {current_quarter_end}
    - "Previous year's quarter" refers to: {prev_year_quarter_start} to {prev_year_quarter_end}
    - "Last quarter 2024" specifically refers to: {current_quarter_start} to {current_quarter_end}
    - "Last quarter 2023" specifically refers to: Q4 of FY 2023-2024, which is January 1, 2024 to March 31, 2024
    
    IMPORTANT TERMINOLOGY RULES:
    1. "Last quarter" without specifying a year ALWAYS refers to the current quarter we are in right now.
       EXAMPLE: Today is {current_date}, so "last quarter" means {current_quarter_start} to {current_quarter_end}.
    
    2. "Last quarter of YYYY" or "last quarter YYYY" ALWAYS refers to Q4 of FY YYYY-(YYYY+1), which is January-March of (YYYY+1).
       EXAMPLE: "Last quarter of 2024" MUST be interpreted as January 1, 2025 to March 31, 2025.
       EXAMPLE: "Last quarter of 2023" MUST be interpreted as January 1, 2024 to March 31, 2024.
    
    3. When comparing to "previous year", ALWAYS compare the same quarter of the previous financial year.
       EXAMPLE: If we're looking at {current_quarter_start} to {current_quarter_end}, 
       then "previous year" means {prev_year_quarter_start} to {prev_year_quarter_end}.
    
    4. SAP data in this organization is organized by financial year where April is period 1, May is period 2, etc.
   
    For the specific query "{user_question}", you MUST interpret:
    - "Last quarter 2024" as {current_quarter_start} to {current_quarter_end} (Q4 of FY 2024-2025)
    - "Previous year" as {prev_year_quarter_start} to {prev_year_quarter_end} (Q4 of FY 2023-2024)
   
    Transform the user's question into a clear analysis plan using plain text only. Do not use markdown, bullets, or numbering. Format your response as follows:
    
    INVENTORY TURNOVER RATIO:
    - The Inventory Turnover Ratio is calculated as:
      **Total Opening Stock / Total Consumption**
    - Total Opening Stock is the stock available at the beginning of the period.
    - Total Consumption is the sum of all outflows (deductions).
    - Consider only records where SHKZG = 'H' for total consumption.
    - Ensure data is filtered based on financial periods and correctly aggregated before calculation.

    SQL QUERY REQUIREMENTS:
    - Ensure the query is optimized for large datasets by using appropriate indexing and partitioning techniques.
    - Use INNER JOINs or LEFT JOINs only when necessary to minimize performance overhead.
    - Implement proper filtering based on financial year and date ranges.
    - If aggregation is needed, use GROUP BY and window functions where appropriate.
    - Ensure data integrity by filtering out NULL values where necessary.
    - Use CTEs (Common Table Expressions) to break down complex queries into manageable parts.
    - Convert all date fields into 'YYYY-MM-DD' format before performing operations.
    - If ranking or percentile calculations are required, use the RANK() or PERCENTILE_DISC() functions.
    - When dealing with hierarchical data, use recursive CTEs for optimal performance.
    - Avoid SELECT *, explicitly define the required columns for clarity and efficiency.

    CRITICAL BUSINESS RULES:
    - For consumption calculations, only consider records where SHKZG = 'H'.
    - For procurement or supply calculations, only consider records where SHKZG = 'S'.
    - For inventory calculations, consider both:
        - 'S' = Addition (positive MENGE).
        - 'H' = Deduction (negative MENGE).
    - For invoice number, vendor number, and financial year, use:
        - Invoice Number: EBELN.
        - Vendor Number: LIFNR.
        - Financial Year: BUDAT_MKPF.
    - Convert BUDAT_MKPF from 'YYYYMMDD' to 'YYYY-MM-DD' before performing date-based filtering.

    QUERY RESPONSE FORMAT:

    "Analysis Plan: [Restate the core question in specific terms, clearly specifying the exact date ranges]
    Data needed: [List the specific tables and fields required]
    Time periods: [Define exact date ranges with specific months and years - be very explicit]
    Calculations: [Specify exact formulas for all metrics mentioned]
    Filters: [Define precise filtering conditions]
    Sorting: [Specify the order of results]"
   
    Keep your response simple, clear, and in plain text format. Focus on making vague terms specific and defining precise calculations.
 
    Keep your response simple, clear, and in plain text format. Focus on making vague terms specific and defining precise calculations.
 
    Note: - For consumption calculations, only consider records where SHKZG = 'H'. (Important)
          - For procurement or supply calculations, only consider recods where SHKZG='S'.(Important)
          - For inventory calculation, consider both 'S' = Addition (positive MENGE) and 'H' = Deduction (negative MENGE). (Important)
          - For purchase order no., invoice number, vendor number and financial year take XBLNR_MKPF, EBELN, LIFNR, and BUDAT_MKPF columns from mseg table. (Important)
          - Convert the datatype of BUDAT_MKPF from 'YYYYMMDD' to 'YYYY-MM-DD'. (Important)
          - Do not provide any raw SQL query, only step by step instructions.
    """
        return enhancement_prompt

    def sql_prompt(self, enhanced_question):
        """
        Generate the SQL prompt for OpenAI based on the user question.
        """
        current_date = datetime.now().strftime('%Y-%m-%d')
        sql_prompt = f"""
        You are a highly skilled Spark SQL expert and data analyst specializing in Fabric lakehouse environments. Generate an optimized Spark SQL query strictly using only the tables mseg (Material Document Segment) and mbewh (Material Valuation History) from the database schema.
        Enhanced User Question: {enhanced_question}
        Today's date is: {current_date}

        Table Reference Rules:
        - Columns from mseg:
        - MATNR (Material Number, STRING)
        - BUDAT_MKPF (Posting Date, STRING)
        - SHKZG (Debit/Credit Indicator, STRING)
        - MENGE (Quantity, STRING)
        - LIFNR (Vendor Number, STRING)
        - EBELN (Purchase Order No., STRING)
        - MAA_URZEI (Origin Indicator, STRING)
        - BWART (Movement Type, STRING)
        - WERKS (Plant, STRING)
        - LGORT (Storage Location, STRING)
        - CHARG (Batch Number, STRING)
        - XBLNR_MKPF (Invoice No., STRING)
        - Columns from mbewh:
        - MATNR (Material Number, STRING)
        - LFMON (Financial Month, STRING)
        - LFGJA (Fiscal Year, STRING)
        - LBKUM (Closing Stock, STRING)

        Column Logic:
        - For SHKZG (Debit/Credit Indicator):
        - 'S' = Addition (positive MENGE) that is Procurement quantity or Supply
        - 'H' = Deduction (negative MENGE) that is Consumption
        - String comparisons should use equality operators (= or <>) with proper quoting
        - For string pattern matching, use LIKE operator with appropriate wildcards

        Opening Stock Calculation:
        - Opening Stock for each month = Previous month's LBKUM.
        - Use LFMON - 1 for the previous month.
        - If LFMON = 1 (April), use LFMON = 12 of the previous year (LFGJA - 1).

        Inventory Turn Over Ratio Calculation:
        **Total Opening Stock / Total Consumption**
        - Total Opening Stock is the stock available at the beginning of the period.
        - Total Consumption is the sum of all outflows (deductions).
        - Consider only records where SHKZG = 'H' for total consumption.
        - Ensure data is filtered based on financial periods and correctly aggregated before calculation.

        Calendar to LFMON Conversion:  
        - April = 1, May = 2, June = 3, July = 4, August = 5, September = 6, October = 7, November = 8, December = 9, January = 10, February = 11, March = 12.

        Additional Guidelines:
        - Use the schema name {self.schema} for all table references without square brackets (e.g., schema.mseg instead of [schema].mseg).
        - Use only the from mseg and mbewh tables; no other tables are allowed.
        - Ensure MENGE is adjusted based on SHKZG logic.
        - Dynamically calculate Opening Stock from mbewh.
        - Properly handle all string comparisons by using single quotes around string literals (e.g., SHKZG = 'S', not SHKZG = S)
        - When comparing string values in WHERE clauses or joins, use proper string equality operators (= or <>)
        - For string pattern matching, use LIKE with appropriate wildcards (%, _)
        - For string concatenation, use the concat() function instead of the + operator
        - Handle potential NULL values in string columns appropriately (use IS NULL or IS NOT NULL)
        - Optimize the SQL query for Spark SQL performance:
        - Use appropriate window functions instead of self-joins when possible
        - Minimize subqueries that require shuffling large amounts of data
        - Use proper date functions compatible with Spark SQL
        - Consider data partitioning when filtering on date columns
        - Do not include explanations—return only the SQL query as plain text (no Markdown formatting).
        - Apply all the necessary calculations which enhanced_question [{enhanced_question}] mentioned.
        - Do not use square brackets [] anywhere in the SQL query, as they're not supported in Spark SQL.
        - Use backticks (') if identifiers need to be escaped, rather than square brackets.
        - For consumption and Procurement quantity or Supply refer to column logic.
        - When using CTEs, place them at the beginning of the query using WITH syntax.
        - For date functions, use Spark SQL-compatible functions (date_add, date_sub, etc.).
        - For conditional logic, ensure CASE statements are properly formatted for Spark SQL.
        - Ensure all string comparisons are case-sensitive unless specified otherwise
        - First, inspect the actual column names available in the mseg table schema. Based on the error message, BUDAT_MKPF is not available but there might be columns like BUDAT_MKPF, CPUDT_MKPF, or BUSTM.
        Note:
        - For consumption calculations, only consider records where SHKZG = 'H'. (Important)
        - For procurement or supply calculations, only consider records where SHKZG = 'S'. (Important)
        - For inventory calculation, consider both 'S' = Addition (positive MENGE) and 'H' = Deduction (negative MENGE). (Important)
        - For purchase order number, invoice number, vendor number and financial year take XBLNR_MKPF, EBELN, LIFNR, and BUDAT_MKPF columns from mseg table. (Important)
        - String columns like MATNR, LIFNR, EBELN, CHARG, etc. should be properly quoted in the query for comparison.
        - Financial year (FY), fiscal year, or simply 'year' starts from April 1st of a given year to March 31st of the following year. (Important)
        - Q1 2024 (first Quarter) = April 1, 2024 to June 1, 2024; Q2 2024 (second Quarter) = July 1, 2024 to September 1, 2024. (Important)
        - Q3 2024 (third Quarter) = October 1, 2024 to December 1, 2024; Q4 2024 (last quarter or fourth Quarter) = January 1, 2025 to March 1, 2025. (Important)
        Provide only the SQL query as plain text without any formatting or additional text.
        """
        return sql_prompt
    
    def enhancement_prompt_an(self, user_question, prompt_instructions=""):
        """
        Enhances the user's query by adding relevant schema details, instructions, and the current date.
    
        IMPORTANT:
        - "Last quarter" without a year refers to the current quarter
        - "Last quarter of YYYY" refers to Q4 of FY YYYY-(YYYY+1), which is Jan-Mar of (YYYY+1)
        """
        # Use the current date
        current_date = datetime.now().strftime('%Y-%m-%d')
 
        # Extract year and month from current date
        current_datetime = datetime.now()
        current_year = current_datetime.year
        current_month = current_datetime.month
    
        # Determine the current financial year
        current_fy_start_year = current_year if current_month >= 4 else current_year - 1
        current_fy_end_year = current_fy_start_year + 1
    
        # Determine the current quarter within the financial year
        if 4 <= current_month <= 6:
            current_quarter = 1
        elif 7 <= current_month <= 9:
            current_quarter = 2
        elif 10 <= current_month <= 12:
            current_quarter = 3
        else:  # 1 <= current_month <= 3
            current_quarter = 4
    
        # Define exact date ranges for current and previous year quarters
        if current_quarter == 1:
            current_quarter_start = f"{current_fy_start_year}-04-01"
            current_quarter_end = f"{current_fy_start_year}-06-30"
            prev_year_quarter_start = f"{current_fy_start_year-1}-04-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-06-30"
        elif current_quarter == 2:
            current_quarter_start = f"{current_fy_start_year}-07-01"
            current_quarter_end = f"{current_fy_start_year}-09-30"
            prev_year_quarter_start = f"{current_fy_start_year-1}-07-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-09-30"
        elif current_quarter == 3:
            current_quarter_start = f"{current_fy_start_year}-10-01"
            current_quarter_end = f"{current_fy_start_year}-12-31"
            prev_year_quarter_start = f"{current_fy_start_year-1}-10-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-12-31"
        else:  # current_quarter == 4
            current_quarter_start = f"{current_fy_end_year}-01-01"
            current_quarter_end = f"{current_fy_end_year}-03-31"
            prev_year_quarter_start = f"{current_fy_end_year-1}-01-01"
            prev_year_quarter_end = f"{current_fy_end_year-1}-03-31"
    
        enhancement_prompt_an = f"""
    You are a SQL expert with deep knowledge of anomaly detection in structured datasets. Your goal is to refine 
    and enhance user queries based on the mseg_anomalies_ctas table structure while ensuring proper handling 
    of anomaly-related conditions.
   
    Today's Date: {current_date}
    Schema: {self.schema}
    Prompt Instructions: {prompt_instructions}
    User Question: {user_question}

    Table Structure:
    - {self.schema}.mseg_anomalies_ctas contains multiple anomaly detection columns.
    - Each attribute 'X' has two corresponding anomaly columns:
        1. 'X_anomaly' (binary flag where 1 = anomaly detected, 0 = normal)
        2. 'X_anomaly_score' (numerical score indicating anomaly strength)
    
    ANOMALY DETECTION RULES (CRITICAL):
    - An anomaly is identified by 'X_anomaly' = 1
    - The 'X_anomaly_score' indicates the strength of the anomaly:
        * POSITIVE values (> 0) indicate stronger anomalies
        * NEGATIVE values (< 0) indicate normal behavior
        * Higher positive scores indicate more significant anomalies
    
    - To find the most significant anomalies, sort by 'X_anomaly_score' in DESCENDING order
    - The only table available for querying is {self.schema}.mseg_anomalies_ctas.

    Query Transformation Guide:
    - Identify Key Attributes: Extract relevant columns from the user request from the {self.schema}.mseg_anomalies_ctas table only.
    - Apply Anomaly Filters: Use 'X_anomaly'=1 to identify anomalies from the {self.schema}.mseg_anomalies_ctas table.
    - Filter Conditions: Include any user-defined filtering criteria (e.g., date range, specific materials).
    - Sorting Rules: Ensure high anomalies appear first using ORDER BY 'X_anomaly_score' DESC.
    - Table Referencing: Ensure queries reference '{self.schema}.mseg_anomalies_ctas' correctly.

    Transform the user's question into a clear analysis plan using plain text only. Do not use markdown, bullets, or numbering.
    
    CRITICAL INFORMATION ABOUT DATE INTERPRETATIONS:
   
    The organization follows an April-March financial year structure:
    - Financial Year X-Y runs from April 1 of year X to March 31 of year Y
    - The current financial year is FY {current_fy_start_year}-{current_fy_end_year}
    - The current quarter is Q{current_quarter} of FY {current_fy_start_year}-{current_fy_end_year}
   
    The quarters in each financial year are defined as:
    - Q1: April 1 to June 30
    - Q2: July 1 to September 30
    - Q3: October 1 to December 31
    - Q4: January 1 to March 31
   
    EXPLICIT DATE MAPPINGS FOR TODAY ({current_date}):
    - "Last quarter" or "current quarter" refers to: {current_quarter_start} to {current_quarter_end}
    - "Previous year's quarter" refers to: {prev_year_quarter_start} to {prev_year_quarter_end}
    - "Last quarter 2024" specifically refers to: {current_quarter_start} to {current_quarter_end}
    - "Last quarter 2023" specifically refers to: Q4 of FY 2023-2024, which is January 1, 2024 to March 31, 2024
    
    IMPORTANT TERMINOLOGY RULES:
    1. "Last quarter" without specifying a year ALWAYS refers to the current quarter we are in right now.
       EXAMPLE: Today is {current_date}, so "last quarter" means {current_quarter_start} to {current_quarter_end}.
    
    2. "Last quarter of YYYY" or "last quarter YYYY" ALWAYS refers to Q4 of FY YYYY-(YYYY+1), which is January-March of (YYYY+1).
       EXAMPLE: "Last quarter of 2024" MUST be interpreted as January 1, 2025 to March 31, 2025.
       EXAMPLE: "Last quarter of 2023" MUST be interpreted as January 1, 2024 to March 31, 2024.
    
    3. When comparing to "previous year", ALWAYS compare the same quarter of the previous financial year.
       EXAMPLE: If we're looking at {current_quarter_start} to {current_quarter_end}, 
       then "previous year" means {prev_year_quarter_start} to {prev_year_quarter_end}.
    
    4. SAP data in this organization is organized by financial year where April is period 1, May is period 2, etc.
   
    For the specific query "{user_question}", you MUST interpret:
    - "Last quarter 2024" as {current_quarter_start} to {current_quarter_end} (Q4 of FY 2024-2025)
    - "Previous year" as {prev_year_quarter_start} to {prev_year_quarter_end} (Q4 of FY 2023-2024)
   
    Transform the user's question into a clear analysis plan using plain text only. Do not use markdown, bullets, or numbering. Format your response as follows:
 
    "Analysis Plan: [Restate the core question in specific terms, clearly specifying the exact date ranges]
    Data needed: [List the specific tables and fields required]
    Anomaly Handling: [Specify which anomaly fields to use, e.g., 'X_anomaly=1 AND X_anomaly_score > 0.1' or the threshold specified by the user]
    Time periods: [Define exact date ranges with specific months and years - be very explicit]
    Filters: [Apply 'X_anomaly=1 AND X_anomaly_score > 0.1' or the user-specified threshold, plus any additional filters]
    Sorting: [Specify the order of results - typically 'X_anomaly_score DESC' to show most significant anomalies first]"
   
    Keep your response simple, clear, and in plain text format. Focus on making vague terms specific and defining precise calculations.
 
    Note: 
    - When filtering anomalies, ALWAYS use BOTH conditions: 'X_anomaly = 1' AND 'X_anomaly_score > threshold' (default 0.1 unless user specifies otherwise)
    - Remember that positive scores (> 0) indicate anomalies, with higher scores showing stronger anomalies 
    - Negative scores (< 0) indicate normal behavior (not anomalies)
    - If the user wants the highest anomalies, sort results in descending order of 'X_anomaly_score'
    - Convert the datatype of BUDAT_MKPF from 'YYYYMMDD' to 'YYYY-MM-DD' in {self.schema}.mseg_anomalies_ctas
    - For quantity, plant, amount, moving indicator, invoice no., vendor no. use MENGE, WERKS, DMBTR, BWART, XBLNR_MKPF, LIFNR from mseg_anomalies_ctas table
    - If the generated SQL query includes BUDAT_MKPF, it should always refer to the column BUDAT_MKPF
    - Detect anomalies in transactional data by identifying unusual patterns based on key attributes, focusing on records where X_anomaly=1 AND X_anomaly_score > 0.1 (or the user-specified threshold)

    """
        return enhancement_prompt_an
    
    def sql_prompt_an(self, enhanced_question):
        """
        Generate an optimized Spark SQL query based on an enhanced user input using Azure OpenAI GPT-4o.
        """
        current_date = datetime.now().strftime('%Y-%m-%d')
        sql_prompt_an = f"""
        You are a highly skilled SQL expert specializing in *Spark SQL within Microsoft Fabric*. Your task is to generate a 
        fully optimized SQL query based on the user's request while ensuring proper anomaly detection and handling. 
        Enhanced User Question: {enhanced_question}
        Today's date is: {current_date}

        Table Reference Rules:
        - Columns from 'mseg_anomalies_ctas':
        -  MATNR (Material Number, STRING)
        -  BUDAT_MKPF (Posting Date, STRING)
        -  DMBTR (Amount, STRING) 
        -  MENGE (Quantity, STRING)
        -  BWART (moving_indicator, STRING)
        -  WERKS (Plant, STRING)
        -  XBLNR_MKPF (invoice no., STRING)
        -  LIFNR (vendor no., STRING)
        -  MENGE_anomaly (Quantity Anomaly, INT/BOOLEAN)
        -  MENGE_anomaly_score (Quantity Anomaly Score, FLOAT)
        -  WERKS_anomaly (Plant Anomaly, INT/BOOLEAN)
        -  WERKS_anomaly_score (Plant Anomaly Score, FLOAT)
        -  BWART_anomaly (Movement Type Anomaly, INT/BOOLEAN)
        -  BWART_anomaly_score (Movement Type Anomaly Score, FLOAT)
        -  DMBTR_anomaly (Amount Anomaly, INT/BOOLEAN)
        -  DMBTR_anomaly_score (Amount Anomaly Score, FLOAT)
        -  XBLNR_MKPF_anomaly (invoice no. Anomaly, LONG)
        -  XBLNR_MKPF_anomaly_score (invoice no. Anomaly Score, FLOAT)
        -  LIFNR_anomaly (vendor no. Anomaly, LONG)
        -  LIFNR_anomaly_score (vendor no. Anomaly Score, FLOAT)
    
        Query Context:
        The query will retrieve insights from the '{self.schema}.mseg_anomalies_ctas' table, which contains anomaly detection columns
        for various attributes. Each attribute 'X' has two corresponding anomaly columns:  
    
        1. A binary flag ('X_anomaly'):
        - Value of 1 indicates an anomaly
        - Value of 0 means normal data (not an anomaly)
    
        2. A numeric score ('X_anomaly_score'):
        - Measures the strength/significance of the anomaly
        - Higher positive scores indicate stronger anomalies
        - Negative scores indicate normal behavior
        - Used for sorting and ranking anomalies

        Anomaly Handling Logic (CRITICAL):
        - When filtering for anomalies, use 'X_anomaly' = 1
    
        - For sorting the most significant anomalies, use ORDER BY 'X_anomaly_score' DESC
        - If multiple attributes are involved, apply AND conditions to filter anomalies across them
        - Ensure NULL safety: Handle cases where 'X_anomaly' or 'X_anomaly_score' may be NULL using appropriate IS NOT NULL checks
        - When selecting top anomalies, use LIMIT or ROW_NUMBER() to retrieve the specified number of records

        Additional Guidelines:
        - Use the schema name '{self.schema}' for all table references
        - Query ONLY columns from the '{self.schema}.mseg_anomalies_ctas' table - do not reference any other table
        - Optimize the SQL query for Spark SQL performance:
            * Use appropriate window functions instead of self-joins when possible
            * Minimize subqueries that require shuffling large amounts of data
            * Consider data partitioning when filtering on date columns
        - Return only the SQL query as plain text without explanations
        - Apply all necessary calculations as mentioned in the enhanced question
        - Do not use square brackets '[]', use backticks '`' for escaping identifiers if needed
        - Use 'WITH' for CTEs at the beginning of the query when appropriate
        - For date functions, use Spark SQL-compatible functions (date_add, date_sub, etc.)
        - For conditional logic, ensure CASE statements are properly formatted for Spark SQL
        - If multiple anomaly scores are involved, consider using COALESCE for combined sorting: 
        ORDER BY COALESCE(X1_anomaly_score, X2_anomaly_score) DESC
    
        Important Notes:
        - Convert BUDAT_MKPF (BIGINT) to a DATE using: TO_DATE(CAST(BUDAT_MKPF AS STRING), 'yyyyMMdd')
        - Always refer to the date column as BUDAT_MKPF (with the underscore) not as BUDAT_MKPF
        - For quantity, plant, amount, moving indicator, invoice no., vendor no. use 
        MENGE, WERKS, DMBTR, BWART, XBLNR_MKPF, LIFNR from mseg_anomalies_ctas table
        - Remember: 'X_anomaly' = 1 identifies anomalies; 'X_anomaly_score' indicates anomaly strength
        - Provide ONLY the SQL query without any explanations or markdown formatting
        """
        return sql_prompt_an


from datetime import datetime
 
class QAmodels:
    def __init__(self, schema):
        self.schema = schema  # Store schema as an instance attribute
 
    def enhancement_prompt(self, user_question, prompt_instructions=""):
        """
        Enhances the user's query by adding relevant schema details, instructions, and the current date.
        IMPORTANT:
        - "Last quarter" without a year refers to the current quarter
        - "Last quarter of YYYY" refers to Q4 of FY YYYY-(YYYY+1), which is Jan-Mar of (YYYY+1)
        """
        # Use the current date
        current_date = datetime.now().strftime('%Y-%m-%d')
        # Extract year and month from current date
        current_datetime = datetime.now()
        current_year = current_datetime.year
        current_month = current_datetime.month
        # Determine the current financial year
        current_fy_start_year = current_year if current_month >= 4 else current_year - 1
        current_fy_end_year = current_fy_start_year + 1
        # Determine the current quarter within the financial year
        if 4 <= current_month <= 6:
            current_quarter = 1
        elif 7 <= current_month <= 9:
            current_quarter = 2
        elif 10 <= current_month <= 12:
            current_quarter = 3
        else:  # 1 <= current_month <= 3
            current_quarter = 4
        # Define exact date ranges for current and previous year quarters
        if current_quarter == 1:
            current_quarter_start = f"{current_fy_start_year}-04-01"
            current_quarter_end = f"{current_fy_start_year}-06-30"
            prev_year_quarter_start = f"{current_fy_start_year-1}-04-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-06-30"
        elif current_quarter == 2:
            current_quarter_start = f"{current_fy_start_year}-07-01"
            current_quarter_end = f"{current_fy_start_year}-09-30"
            prev_year_quarter_start = f"{current_fy_start_year-1}-07-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-09-30"
        elif current_quarter == 3:
            current_quarter_start = f"{current_fy_start_year}-10-01"
            current_quarter_end = f"{current_fy_start_year}-12-31"
            prev_year_quarter_start = f"{current_fy_start_year-1}-10-01"
            prev_year_quarter_end = f"{current_fy_start_year-1}-12-31"
        else:  # current_quarter == 4
            current_quarter_start = f"{current_fy_end_year}-01-01"
            current_quarter_end = f"{current_fy_end_year}-03-31"
            prev_year_quarter_start = f"{current_fy_end_year-1}-01-01"
            prev_year_quarter_end = f"{current_fy_end_year-1}-03-31"
 
        enhancement_prompt = f"""
        You are a highly skilled SQL expert and data analyst. Your task is to enhance a given user question
        by making it more structured, detailed, and aligned with the provided schema and prompt instructions.
        Today's Date: {current_date}
        Schema: {self.schema}
        Prompt Instructions: {prompt_instructions}
        User Question: {user_question}
        CRITICAL INFORMATION ABOUT DATE INTERPRETATIONS:
        The organization follows an April-March financial year structure:
        - Financial Year X-Y runs from April 1 of year X to March 31 of year Y
        - The current financial year is FY {current_fy_start_year}-{current_fy_end_year}
        - The current quarter is Q{current_quarter} of FY {current_fy_start_year}-{current_fy_end_year}
        The quarters in each financial year are defined as:
        - Q1: April 1 to June 30
        - Q2: July 1 to September 30
        - Q3: October 1 to December 31
        - Q4: January 1 to March 31
        EXPLICIT DATE MAPPINGS FOR TODAY ({current_date}):
        - "Last quarter" or "current quarter" refers to: {current_quarter_start} to {current_quarter_end}
        - "Previous year's quarter" refers to: {prev_year_quarter_start} to {prev_year_quarter_end}
        - "Last quarter 2024" specifically refers to: {current_quarter_start} to {current_quarter_end}
        - "Last quarter 2023" specifically refers to: Q4 of FY 2023-2024, which is January 1, 2024 to March 31, 2024
        IMPORTANT TERMINOLOGY RULES:
        1. "Last quarter" without specifying a year ALWAYS refers to the current quarter we are in right now.
        EXAMPLE: Today is {current_date}, so "last quarter" means {current_quarter_start} to {current_quarter_end}.
        2. "Last quarter of YYYY" or "last quarter YYYY" ALWAYS refers to Q4 of FY YYYY-(YYYY+1), which is January-March of (YYYY+1).
        EXAMPLE: "Last quarter of 2024" MUST be interpreted as January 1, 2025 to March 31, 2025.
        EXAMPLE: "Last quarter of 2023" MUST be interpreted as January 1, 2024 to March 31, 2024.
        3. When comparing to "previous year", ALWAYS compare the same quarter of the previous financial year.
        EXAMPLE: If we're looking at {current_quarter_start} to {current_quarter_end},
        then "previous year" means {prev_year_quarter_start} to {prev_year_quarter_end}.
        4. SAP data in this organization is organized by financial year where April is period 1, May is period 2, etc.
       
        DATA CONVERSION RULES:
        1. ALWAYS convert BUDAT_MKPF to date using this EXACT syntax:
        CAST(CONVERT(VARCHAR(8), BUDAT_MKPF, 112) AS DATE)
        2. NEVER use DATEADD function with BUDAT_MKPF as an argument (Important and critical).
        3. MSEG.BUDAT_MKPF is stored in a format that requires this specific conversion to be handled properly.
        4. Do not attempt to convert dates using the Unix epoch or seconds-based approaches.
       
        For the specific query "{user_question}", you MUST interpret:
        - "Last quarter 2024" as {current_quarter_start} to {current_quarter_end} (Q4 of FY 2024-2025)
        - "Previous year" as {prev_year_quarter_start} to {prev_year_quarter_end} (Q4 of FY 2023-2024)
        Transform the user's question into a clear analysis plan using plain text only. Do not use markdown, bullets, or numbering. Format your response as follows:
       
        Outlier Calculation:
        "Generate an optimized SQL query to find outliers in the MENGE column.
        The outliers are defined using the Interquartile Range (IQR) method for both columns:
 
        Formula for Outliers:
        Q1 (First Quartile) = 25th percentile of the column
        Q3 (Third Quartile) = 75th percentile of the column
        IQR (Interquartile Range) = Q3 - Q1
        Lower Bound = Q1 - (1.5 × IQR)
        Upper Bound = Q3 + (1.5 × IQR)
        Any value in MENGE that is less than the Lower Bound or greater than the Upper Bound is considered an outlier or anomaly.
        Use the PERCENTILE_CONT function to calculate Q1 and Q3 efficiently and return only the outlier values from both columns.""Analysis Plan: [Restate the core question in specific terms, clearly specifying the exact date ranges]
       
        Data needed: [List the specific tables and fields required]
        Time periods: [Define exact date ranges with specific months and years - be very explicit]
        Calculations: [Specify exact formulas for all metrics mentioned]
        Filters: [Define precise filtering conditions]
        Sorting: [Specify the order of results]"
        Keep your response simple, clear, and in plain text format. Focus on making vague terms specific and defining precise calculations.
        Note: - For consumption calculations, only consider records where SHKZG = 'H'. (Important)
            - For procurement or supply calculations, only consider recods where SHKZG='S'.(Important)
            - For inventory calculation, consider both 'S' = Addition (positive MENGE) and 'H' = Deduction (negative MENGE). (Important)
            - For invoice number, purchase order number, vendor number and financial year take XBLNR_MKPF, EBELN, LIFNR, and BUDAT_MKPF columns from mseg table. (Important)
            - If a user asks about duplicate invoice number always filter BWART = 105. (Important)
            - Any value in MENGE that is less than the Lower Bound or greater than the Upper Bound is considered an outlier or anomaly. (Important)
            - For outlier or anomaly calculation strictly take only year or time period which is mentioned in the query. (Important)
            - For outlier or anomaly calculation if the user mentioned 2024 in the query do not compare with previous year and take the financial year. (Important)
            - For outlier or anomaly calculation if the user mention vendor number then only take LIFNR column from mseg table. (Important)
            - Generate an SQL Server-compatible query for BUDAT_MKPF, avoiding incorrect default dates like 1970-01-01. (Important)
            - Generate an SQL Server-compatible query for BUDAT_MKPF, avoiding incorrect default dates like 1900-01-01. (Important)
            - Convert the datatype of BUDAT_MKPF from 'YYYYMMDD' to 'YYYY-MM-DD'. (Important)
        """
        return enhancement_prompt
 
    def sql_prompt(self, enhanced_question):
        """
        Generate the SQL prompt for OpenAI based on the user question.
        """
        current_date = datetime.now().strftime('%Y-%m-%d')
        sql_prompt = f"""
        You are a highly skilled SQL expert and data analyst specializing in Fabric SQL database environments. Generate an optimized SQL Server-compatible query strictly using only the tables mseg (Material Document Segment) and mbewh (Material Valuation History) from the database schema.
        Enhanced User Question: {enhanced_question}
        Today's date is: {current_date}
 
        Table Reference Rules:
        - Columns from mseg:
        - MATNR (Material Number, nvarchar(max))
        - BUDAT_MKPF (Posting Date, bigint)
        - SHKZG (Debit/Credit Indicator, nvarchar(max))
        - MENGE (Quantity, float)
        - LIFNR (Vendor Number, nvarchar(max))
        - EBELN (Purchase Order No., bigint)
        - MAA_URZEI (Origin Indicator, bigint)
        - BWART (Movement Type, bigint)
        - WERKS (Plant, bigint)
        - LGORT (Storage Location, nvarchar(max))
        - CHARG (Batch Number, nvarchar(max))
        - DMBTR (Amount, float)
        - XBLNR_MKPF (Invoice Number, nvarchar(max))
        - Columns from mbewh:
        - MATNR (Material Number, nvarchar(max))
        - LFMON (Financial Month, bigint)
        - LFGJA (Fiscal Year, bigint)
        - LBKUM (Closing Stock, float)
 
        Column Logic:
        - For SHKZG (Debit/Credit Indicator):
        - 'S' = Addition (positive MENGE) that is Procurement quantity or Supply
        - 'H' = Deduction (negative MENGE) that is Consumption
        - String comparisons should use equality operators (= or <>) with proper quoting
        - For string pattern matching, use LIKE operator with appropriate wildcards
 
        Opening Stock Calculation:
        - Opening Stock for each month = Previous month's LBKUM.
        - Use LFMON - 1 for the previous month.
        - If LFMON = 1 (April), use LFMON = 12 of the previous year (LFGJA - 1).
       
        Outlier Calculation:
        "Generate an optimized SQL query to find outliers in the MENGE column.
        The outliers are defined using the Interquartile Range (IQR) method for both columns:
 
        Formula for Outliers:
        Q1 (First Quartile) = 25th percentile of the column
        Q3 (Third Quartile) = 75th percentile of the column
        IQR (Interquartile Range) = Q3 - Q1
        Lower Bound = Q1 - (1.5 × IQR)
        Upper Bound = Q3 + (1.5 × IQR)
        Any value in MENGE that is less than the Lower Bound or greater than the Upper Bound is considered an outlier or anomaly.
        Use the PERCENTILE_CONT function to calculate Q1 and Q3 efficiently and return only the outlier values from both columns."
 
        Calendar to LFMON Conversion:  
        - April = 1, May = 2, June = 3, July = 4, August = 5, September = 6, October = 7, November = 8, December = 9, January = 10, February = 11, March = 12.
 
        Additional Guidelines:
        - Use the schema name {self.schema} for all table references without square brackets (e.g., schema.mseg instead of [schema].mseg).
        - Use only the mseg and mbewh tables; no other tables are allowed.
        - Ensure MENGE is adjusted based on SHKZG logic.
        - Dynamically calculate Opening Stock from mbewh.
        - Properly handle all string comparisons by using single quotes around string literals (e.g., SHKZG = 'S', not SHKZG = S)
        - When comparing string values in WHERE clauses or joins, use proper string equality operators (= or <>)
        - For string pattern matching, use LIKE with appropriate wildcards (%, _)
        - For string concatenation, use the concat() function instead of the + operator
        - Handle potential NULL values in string columns appropriately (use IS NULL or IS NOT NULL)
        - Optimize the SQL query for SQL Server performance:
        - Use appropriate window functions instead of self-joins when possible
        - Minimize subqueries that require shuffling large amounts of data
        - Use proper date functions compatible with SQL Server
        - Consider data partitioning when filtering on date columns
        - Do not include explanations—return only the SQL query as plain text (no Markdown formatting).
        - Apply all the necessary calculations which enhanced_question [{enhanced_question}] mentioned.
        - Do not use square brackets [] anywhere in the SQL Server-compatible query, as they're not supported in SQL Server.
        - Use backticks (`) if identifiers need to be escaped, rather than square brackets.
        - For consumption and Procurement quantity or Supply refer to column logic.
        - When using CTEs, place them at the beginning of the query using WITH syntax.
        - For date functions, use SQL Server-compatible functions (DATEADD).
        - For conditional logic, ensure CASE statements are properly formatted for SQL Server.
        - Ensure all string comparisons are case-sensitive unless specified otherwise
        - First, inspect the actual column names available in the mseg table schema. Based on the error message, BUDAT_MKPF is not available but there might be columns like BUDAT_MKPF, CPUDT_MKPF, or BUSTM.
        Note:
        - For consumption calculations, only consider records where SHKZG = 'H'. (Important)
        - For procurement or supply calculations, only consider records where SHKZG = 'S'. (Important)
        - For inventory calculation, consider both 'S' = Addition (positive MENGE) and 'H' = Deduction (negative MENGE). (Important)
        - For invoice number, purchase order number, vendor number and financial year take XBLNR_MKPF, EBELN, LIFNR, and BUDAT_MKPF columns from mseg table. (Important)
        - String columns like MATNR, LIFNR, EBELN, CHARG, etc. should be properly quoted in the query for comparison.
        - Financial year (FY), fiscal year, or simply 'year' starts from April 1st of a given year to March 31st of the following year. (Important)
        - Q1 2024 (first Quarter) = April 1, 2024 to June 1, 2024; Q2 2024 (second Quarter) = July 1, 2024 to September 1, 2024. (Important)
        - Q3 2024 (third Quarter) = October 1, 2024 to December 1, 2024; Q4 2024 (last quarter or fourth Quarter) = January 1, 2025 to March 1, 2025. (Important)
        - Generate an SQL Server-compatible query for BUDAT_MKPF, avoiding incorrect default dates like 1970-01-01. (Important)
        - Generate an SQL Server-compatible query for BUDAT_MKPF, avoiding incorrect default dates like 1900-01-01. (Important)
        - If a user asks about duplicate invoice number always filter BWART = 105. (Important)
        - Any value in MENGE that is less than the Lower Bound or greater than the Upper Bound is considered an outlier or anomaly. (Important)
        Provide only the SQL Server-compatible query as plain text without any formatting or additional text.
 
        """
        return sql_prompt


# -- coding: utf-8 --
"""
Created on Fri oct  10 16:15:41 2023

@author: Shuvadeep
"""
from datetime import datetime as dt    
import numpy as np
import pandas as pd

class cust_seg:
    def __init__(self, df,today_date):
        self.df = df
        self.today_date=today_date
    
    def Daily (self):
        df=self.df
        # Convert data types of specific columns to strings
        df[["VISIT_ID", "SKU_ID", "SKU_DESC", "DEPT_DESC"]] = df[["VISIT_ID", "SKU_ID", "SKU_DESC", "DEPT_DESC"]].astype("string")
        # Convert CUST_ID to integer
        df["CUST_ID"] = df["CUST_ID"].astype(int)
        # Defining today date as max(InvoiceDate) + 2 days
        #date_input = input("Enter a date in the format YYYY-MM-DD: ")
        today_date = dt.strptime(self.today_date, "%Y-%m-%d")
        df["VISIT_DT"]=pd.to_datetime(df["VISIT_DT"],format='%Y-%m-%d')
        df['week']=pd.PeriodIndex(df.VISIT_DT, freq='W')
        df['day']=pd.PeriodIndex(df.VISIT_DT, freq='D')
        df_pivot = self.df.pivot_table(index='CUST_ID', columns='week', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        # Format pivot table
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        perc_25 = np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25 = pd.DataFrame()
        df_perc_25["CUST_ID"] = df_pivot.index
        df_perc_25["Perc_25"] = perc_25
        df_daily = df_perc_25[df_perc_25["Perc_25"] >1.0]
        df_daily.drop(columns = {"Perc_25"}, axis=1, inplace=True)
        #RFM Calculation 
        rfm = df.groupby("CUST_ID").agg({"VISIT_DT":lambda date: (today_date - date.max()).days,
                                    "VISIT_ID": lambda num: num.nunique(),
                                    "SALES": lambda price: price.sum()}) #total price per customer
        rfm.columns = ['Recency', 'Frequency', "Monetary"]
        rfm_daily=rfm[rfm["Recency"]<30]
        self.df_daily_active= rfm_daily[rfm_daily.index.isin(df_daily["CUST_ID"])]
        rfm_daily_inactive=rfm[rfm["Recency"]>30]
        df_daily_inactive = rfm_daily_inactive[rfm_daily_inactive.index.isin(df_daily.CUST_ID)]
        q2_F=self.df_daily_active["Frequency"].quantile(0.50)
        q2_M = self.df_daily_active["Monetary"].quantile(0.50)
        # create a list of our conditions
        conditions = [
            (self.df_daily_active['Frequency'] <= q2_F) & (self.df_daily_active['Monetary']<=q2_M),
            (self.df_daily_active['Frequency'] <= q2_F) & (self.df_daily_active['Monetary']>q2_M),
            (self.df_daily_active['Frequency']>q2_F) & (self.df_daily_active['Monetary']<=q2_M),
            (self.df_daily_active['Frequency']>q2_F) & (self.df_daily_active['Monetary']>q2_M)
            ]
        # create a list of the values we want to assign for each condition
        values = ['LFLM', 'LFHM', 'HFLM', 'HFHM']
        # create a new column and use np.select to assign values to it using our lists as arguments
        self.df_daily_active['Customer_segments'] = np.select(conditions, values)
        df_daily_active_fn=self.df_daily_active.groupby('Customer_segments').median().rename(columns= {'Recency':"Median Recency","Frequency":"Median Frequency","Monetary":'Median Monetary'}).join(self.df_daily_active.Customer_segments.value_counts().rename('count'))
        return rfm_daily.round(2),self.df_daily_active.round(2),df_daily_inactive.round(2),df_daily_active_fn.round(2)
    
    def Weekly (self):
        df=self.df
        # Convert data types of specific columns to strings
        df[["VISIT_ID", "SKU_ID", "SKU_DESC", "DEPT_DESC"]] = df[["VISIT_ID", "SKU_ID", "SKU_DESC", "DEPT_DESC"]].astype("string")
        # Convert CUST_ID to integer
        df["CUST_ID"] = df["CUST_ID"].astype(int)
        # Defining today date as max(InvoiceDate) + 2 days
        #date_input = input("Enter a date in the format YYYY-MM-DD: ")
        today_date = dt.strptime(self.today_date, "%Y-%m-%d")
        df["VISIT_DT"]=pd.to_datetime(df["VISIT_DT"],format='%Y-%m-%d')
        df['month']=pd.PeriodIndex(df.VISIT_DT, freq='M')
        df['week']=pd.PeriodIndex(df.VISIT_DT, freq='W')
        df['day']=pd.PeriodIndex(df.VISIT_DT, freq='D')
        df_pivot = self.df.pivot_table(index='CUST_ID', columns='week', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        # Format pivot table
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        perc_25 = np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame(perc_25).reset_index()
        df_perc_25.columns=["Customer ID","Perc_25"]
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25 = pd.DataFrame()     
        df_perc_25["CUST_ID"] = df_pivot.index  
        df_perc_25["Perc_25"] = perc_25    
        df_nondaily = df_perc_25[df_perc_25["Perc_25"] <=1.0] 
        #df_daily = df_perc_25[df_perc_25["Perc_25"] >=2.00]
        df_sub = df[df['CUST_ID'].isin(df_nondaily["CUST_ID"])]
        df_pivot = df_sub.pivot_table(index='CUST_ID', columns='month', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        # 25th percentile of number of purchase days for each customer
        perc_25=np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame()
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25["Perc_25"]=perc_25
        df_weekly = df_perc_25[df_perc_25["Perc_25"] >= 2.0]
        #RFM Calculation 
        rfm = df.groupby("CUST_ID").agg({"VISIT_DT":lambda date: (today_date - date.max()).days,
                                    "VISIT_ID": lambda num: num.nunique(),
                                    "SALES": lambda price: price.sum()}) #total price per customer
        rfm.columns = ['Recency', 'Frequency', "Monetary"]
        df_weekly.drop(columns = {"Perc_25"}, axis=1, inplace=True)
        rfm_weekly_active=rfm[rfm["Recency"]<60]
        df_weekly_active= rfm_weekly_active[rfm_weekly_active.index.isin(df_weekly["CUST_ID"])]
        rfm_weekly_inactive=rfm[rfm["Recency"]>=60]
        df_weekly_inactive = rfm_weekly_inactive[rfm_weekly_inactive.index.isin(df_weekly.CUST_ID)]
        q2_F=df_weekly_active["Frequency"].quantile(0.50)
        q2_M = df_weekly_active["Monetary"].quantile(0.50)
        # create a list of our conditions
        conditions = [
            (df_weekly_active['Frequency'] <= q2_F) & (df_weekly_active['Monetary']<=q2_M),
            (df_weekly_active['Frequency'] <= q2_F) & (df_weekly_active['Monetary']>q2_M),
            (df_weekly_active['Frequency']>q2_F) & (df_weekly_active['Monetary']<=q2_M),
            (df_weekly_active['Frequency']>q2_F) & (df_weekly_active['Monetary']>q2_M)
            ]
        # create a list of the values we want to assign for each condition
        values = ['LFLM', 'LFHM', 'HFLM', 'HFHM']
        # create a new column and use np.select to assign values to it using our lists as arguments
        df_weekly_active['Customer_segments'] = np.select(conditions, values)
        df_weekly_active_fn=df_weekly_active.groupby('Customer_segments').median().rename(columns= {'Recency':"Median Recency","Frequency":"Median Frequency","Monetary":'Median Monetary'}).join(df_weekly_active.Customer_segments.value_counts().rename('count'))
        return rfm_weekly_active.round(2),df_weekly_active.round(2),df_weekly_inactive.round(2),df_weekly_active_fn.round(2)
    
    def Monthly (self):
        df=self.df
        # Convert data types of specific columns to strings
        df[["VISIT_ID", "SKU_ID", "SKU_DESC", "DEPT_DESC"]] = df[["VISIT_ID", "SKU_ID", "SKU_DESC", "DEPT_DESC"]].astype("string")
        # Convert CUST_ID to integer
        df["CUST_ID"] = df["CUST_ID"].astype(int)
        # Defining today date as max(InvoiceDate) + 2 days
        #date_input = input("Enter a date in the format YYYY-MM-DD: ")
        today_date = dt.strptime(self.today_date, "%Y-%m-%d")
        df["VISIT_DT"]=pd.to_datetime(df["VISIT_DT"],format='%Y-%m-%d')
        df['quarter'] = pd.PeriodIndex(df.VISIT_DT, freq='Q')
        df['month']=pd.PeriodIndex(df.VISIT_DT, freq='M')
        df['week']=pd.PeriodIndex(df.VISIT_DT, freq='W')
        df['day']=pd.PeriodIndex(df.VISIT_DT, freq='D')
        df_pivot = self.df.pivot_table(index='CUST_ID', columns='week', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        # Format pivot table
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        perc_25 = np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame(perc_25).reset_index()
        df_perc_25.columns=["Customer ID","Perc_25"]
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25 = pd.DataFrame()     
        df_perc_25["CUST_ID"] = df_pivot.index  
        df_perc_25["Perc_25"] = perc_25    
        df_nondaily = df_perc_25[df_perc_25["Perc_25"] <=1.0] 
        #df_daily = df_perc_25[df_perc_25["Perc_25"] >=2.00]
        df_sub = df[df['CUST_ID'].isin(df_nondaily["CUST_ID"])]
        df_pivot = df_sub.pivot_table(index='CUST_ID', columns='month', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        # 25th percentile of number of purchase days for each customer
        perc_25=np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame()
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25["Perc_25"]=perc_25
        df_nonweekly = df_perc_25[df_perc_25["Perc_25"] < 2.0]
        #df_weekly = df_perc_25[df_perc_25["Perc_25"] >= 2.0]
        df_nonweekly.drop(columns = {"Perc_25"}, axis=1, inplace=True)
        df_sub= df[df['CUST_ID'].isin(df_nonweekly["CUST_ID"])]
        df_pivot = df_sub.pivot_table(index='CUST_ID', columns='quarter', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        perc_25=np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame()
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25["Perc_25"]=perc_25
        df_monthly = df_perc_25[df_perc_25["Perc_25"] >=2.0]
        #RFM Calculation 
        rfm = df.groupby("CUST_ID").agg({"VISIT_DT":lambda date: (today_date - date.max()).days,
                                    "VISIT_ID": lambda num: num.nunique(),
                                    "SALES": lambda price: price.sum()}) #total price per customer

        #rfm.columns = rfm.columns.droplevel(0)
        rfm.columns = ['Recency', 'Frequency', "Monetary"]
        df_monthly.drop(columns = {"Perc_25"}, axis=1, inplace=True)
        rfm_monthly_active=rfm[rfm["Recency"]<90]
        df_monthly_active= rfm_monthly_active[rfm_monthly_active.index.isin(df_monthly["CUST_ID"])]
        rfm_monthly_inactive=rfm[rfm["Recency"]>=90]
        df_monthly_inactive= rfm_monthly_inactive[rfm_monthly_inactive.index.isin(df_monthly.CUST_ID)]
        q2_F=df_monthly_active["Frequency"].quantile(0.50)
        q2_M = df_monthly_active["Monetary"].quantile(0.50)
        # create a list of our conditions
        conditions = [
            (df_monthly_active['Frequency'] <= q2_F) & (df_monthly_active['Monetary']<=q2_M),
            (df_monthly_active['Frequency'] <= q2_F) & (df_monthly_active['Monetary']>q2_M),
            (df_monthly_active['Frequency']>q2_F) & (df_monthly_active['Monetary']<=q2_M),
            (df_monthly_active['Frequency']>q2_F) & (df_monthly_active['Monetary']>q2_M)
            ]
        # create a list of the values we want to assign for each condition
        values = ['LFLM', 'LFHM', 'HFLM', 'HFHM']
        # create a new column and use np.select to assign values to it using our lists as arguments
        df_monthly_active['Customer_segments'] = np.select(conditions, values)
        df_monthly_active_fn=df_monthly_active.groupby('Customer_segments').median().rename(columns={"Recency":"Median Recency","Frequency":"Median Frequency","Monetary":"Median Monetary"}).join(df_monthly_active.Customer_segments.value_counts().rename('count'))
        return rfm_monthly_active.round(2),df_monthly_active.round(2),df_monthly_inactive.round(2),df_monthly_active_fn.round(2)
    
    def Quarterly (self):
        df=self.df
        # Convert data types of specific columns to strings
        df[["VISIT_ID", "SKU_ID", "SKU_DESC", "DEPT_DESC"]] = df[["VISIT_ID", "SKU_ID", "SKU_DESC", "DEPT_DESC"]].astype("string")
        # Convert CUST_ID to integer
        df["CUST_ID"] = df["CUST_ID"].astype(int)
        # Defining today date as max(InvoiceDate) + 2 days
        #date_input = input("Enter a date in the format YYYY-MM-DD: ")
        today_date = dt.strptime(self.today_date, "%Y-%m-%d")
        df["VISIT_DT"]=pd.to_datetime(df["VISIT_DT"],format='%Y-%m-%d')
        df['year']=pd.PeriodIndex(df.VISIT_DT,freq='Y')
        df['quarter'] = pd.PeriodIndex(df.VISIT_DT, freq='Q')
        df['month']=pd.PeriodIndex(df.VISIT_DT, freq='M')
        df['week']=pd.PeriodIndex(df.VISIT_DT, freq='W')
        df['day']=pd.PeriodIndex(df.VISIT_DT, freq='D')
        df_pivot = self.df.pivot_table(index='CUST_ID', columns='week', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        # Format pivot table
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        perc_25 = np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame(perc_25).reset_index()
        df_perc_25.columns=["Customer ID","Perc_25"]
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25 = pd.DataFrame()     
        df_perc_25["CUST_ID"] = df_pivot.index  
        df_perc_25["Perc_25"] = perc_25    
        df_nondaily = df_perc_25[df_perc_25["Perc_25"] <=1.0] 
        #df_daily = df_perc_25[df_perc_25["Perc_25"] >=2.00]
        df_sub = df[df['CUST_ID'].isin(df_nondaily["CUST_ID"])]
        df_pivot = df_sub.pivot_table(index='CUST_ID', columns='month', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        # 25th percentile of number of purchase days for each customer
        perc_25=np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame()
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25["Perc_25"]=perc_25
        df_nonweekly = df_perc_25[df_perc_25["Perc_25"] < 2.0]
        #df_weekly = df_perc_25[df_perc_25["Perc_25"] >= 2.0]
        df_nonweekly.drop(columns = {"Perc_25"}, axis=1, inplace=True)
        df_sub= df[df['CUST_ID'].isin(df_nonweekly["CUST_ID"])]
        df_pivot = df_sub.pivot_table(index='CUST_ID', columns='quarter', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        perc_25=np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame()
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25["Perc_25"]=perc_25
        df_nonmonthly = df_perc_25[df_perc_25["Perc_25"] <2.0]
        df_nonmonthly.drop(columns = {"Perc_25"}, axis=1, inplace=True)
        df_sub= df[df['CUST_ID'].isin(df_nonmonthly["CUST_ID"])]
        df_pivot = df_sub.pivot_table(index='CUST_ID', columns='year', aggfunc={'VISIT_DT': 'nunique'}, fill_value=0)
        df_pivot.columns = df_pivot.columns.droplevel(0)  # remove date
        df_pivot.columns.name = None  # remove week
        # Replace redundant values before the first purchase for each customer
        for i in df_pivot.index:
            for j in df_pivot.columns:
                if df_pivot.loc[i, j] >= 1:
                    break
                else:
                    df_pivot.loc[i][j] = -1
        # Replace -1 (indicating no purchase) with NaN
        df_pivot = df_pivot.replace(-1, np.nan)
        perc_25=np.nanquantile(df_pivot, 0.25, axis=1)
        df_perc_25=pd.DataFrame()
        df_perc_25["CUST_ID"]=df_pivot.index
        df_perc_25["Perc_25"]=perc_25
        df_quarterly = df_perc_25[df_perc_25["Perc_25"] >=3.0]
        #RFM Calculation 
        rfm = df.groupby("CUST_ID").agg({"VISIT_DT":lambda date: (today_date - date.max()).days,
                                    "VISIT_ID": lambda num: num.nunique(),
                                    "SALES": lambda price: price.sum()}) #total price per customer
        rfm.columns = ['Recency', 'Frequency', "Monetary"]
        df_quarterly.drop(columns = {"Perc_25"}, axis=1, inplace=True)
        rfm_quarterly_active=rfm[rfm["Recency"]<180]
        df_quarterly_active = rfm_quarterly_active[rfm_quarterly_active.index.isin(df_quarterly["CUST_ID"])]
        rfm_quarterly_inactive=rfm[rfm["Recency"]>=180]
        df_quarterly_inactive = rfm_quarterly_inactive[rfm_quarterly_inactive.index.isin(df_quarterly["CUST_ID"])]
        q2_F=df_quarterly_active["Frequency"].quantile(0.50)
        q2_M = df_quarterly_active["Monetary"].quantile(0.50)
        # create a list of our conditions
        conditions = [
            (df_quarterly_active['Frequency'] <= q2_F) & (df_quarterly_active['Monetary']<=q2_M),
            (df_quarterly_active['Frequency'] <= q2_F) & (df_quarterly_active['Monetary']>q2_M),
            (df_quarterly_active['Frequency']>q2_F) & (df_quarterly_active['Monetary']<=q2_M),
            (df_quarterly_active['Frequency']>q2_F) & (df_quarterly_active['Monetary']>q2_M)
            ]
        # create a list of the values we want to assign for each condition
        values = ['LFLM', 'LFHM', 'HFLM', 'HFHM']
        # create a new column and use np.select to assign values to it using our lists as arguments
        df_quarterly_active['Customer_segments'] = np.select(conditions, values)
        df_quarterly_active_fn=df_quarterly_active.groupby('Customer_segments').median().rename(columns={"Recency":"Median Recency","Frequency":"Median Frequency","Monetary":"Median Monetary"}).join(df_quarterly_active.Customer_segments.value_counts().rename('count'))
        return rfm_quarterly_active.round(2),df_quarterly_active.round(2),df_quarterly_inactive.round(2),df_quarterly_active_fn.round(2)